from blok.diff import compare_structures, ChangedDiff
from typing import List
from pathlib import Path
import os


class YamlFile:
    """Represents a yaml file"""

    def __init__(self, **values):
        self.values = values

    def __str__(self):
        return str(self.values)

    def diff(self, other: "YamlFile", path: str) -> List[str]:
        return compare_structures(self.values, other.values, path)

    @classmethod
    def is_representable_as(cls, path: Path):
        if path.suffix in [".yaml", ".yml"]:
            return True
        return False


class JSONFile:
    """Represents a yaml file"""

    def __init__(self, **values):
        self.values = values

    def __str__(self):
        return str(self.values)

    def diff(self, other: "YamlFile", path: str) -> List[str]:
        return compare_structures(self.values, other.values, path)

    @classmethod
    def is_representable_as(cls, path: Path):
        if path.suffix in [".json", ".json"]:
            return True
        return False


class Repo:
    """Represents a repository

    Repository will be cloned if it does not exist


    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def diff(self, other: "Repo", path: str) -> List[str]:
        if self.value != other.value:
            return [ChangedDiff(path, self.value, self.value)]
        return []

    def __eq__(self, other):
        if not isinstance(other, Repo):
            return False
        return self.value == other.value

    @classmethod
    def is_representable_as(cls, path: Path):
        if path.is_dir():
            if "__repo__.txt" in os.listdir(path):
                return True

        return False



import os
from pathlib import Path
from blok.tree.models import JSONFile, YamlFile, Repo
import yaml
import json
import subprocess


def create_files_and_folders(base_path, structure, git_command: str = "git"):
    for key, value in structure.items():
        current_path = base_path / key
        if isinstance(value, YamlFile):
            # Create file and write yaml
            with current_path.open("w") as file:
                yaml.dump(value.values, file, yaml.SafeDumper)

        elif isinstance(value, JSONFile):
            # Create file and write yaml
            with current_path.open("w") as file:
                json.dump(value.values, file, indent=2)

        elif isinstance(value, Repo):
            # Create directory and clone repo
            current_path.mkdir(parents=True, exist_ok=True)
            subprocess.run(f"{git_command} clone {value} {current_path}", shell=True)
            # TODO: Clone repo

        elif isinstance(value, dict):
            # Create directory and recurse into it
            current_path.mkdir(parents=True, exist_ok=True)
            create_files_and_folders(current_path, value)
        else:
            # Create file and write bytes if any
            if value is not None:
                if isinstance(value, str):
                    value = value.encode()

                assert isinstance(
                    value, bytes
                ), f"Expected bytes for file content, got {type(value)}"

                with current_path.open("wb") as file:
                    if value is not None:
                        file.write(value)

            else:
                raise ValueError("Value cannot be None")



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass
import socket
from dataclasses import dataclass, field
from typing import Set
from blok.bloks.services.dns import DNSResult, DnsService
from blok.bloks.services.vscode import VsCodeService
from blok.blok import ExecutionContext
from blok.tree.models import JSONFile


@blok(
    VsCodeService,
    options=[Option("with_docker", help="Have a docker up command", default=True)],
)
class VsCodeBlok:
    def __init__(self):
        self.result = None
        self.tasks = []

    def preflight(self, init: InitContext, with_docker=True):
        if with_docker:
            self.register_task(
                "Docker Compose Up",
                "shell",
                "docker",
                ["compose", "up"],
                {"cwd": "${workspaceFolder}"},
            )

    def register_task(
        self,
        label: str,
        type: str,
        command: str,
        args: list[str],
        options: Dict[str, Any],
    ):
        self.tasks.append(
            {
                "label": label,
                "type": type,
                "command": command,
                "args": args,
                "options": options,
            }
        )

    def build(self, context: ExecutionContext):
        context.file_tree.set_nested(
            ".vscode",
            "tasks.json",
            JSONFile(version="2.0.0", tasks=self.tasks),
        )
        pass



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass
import socket
from dataclasses import dataclass, field
from typing import Set
from blok.bloks.services.dns import DNSResult, DnsService
from inquirer import prompt


@dataclass
class DNSResult:
    hostnames: Set[str] = field(default_factory=set)
    ip_addresses: Set[str] = field(default_factory=set)


@blok(
    DnsService,
    options=[
        Option(
            "hostnames",
            help="A list of hostnames to resolve",
            multiple=True,
        ),
        Option(
            "ip_addresses",
            help="A list of IP addresses to resolve",
            multiple=True,
        ),
    ],
    description="Will use the command prompts to get the hostnames.",
)
class PromptDNSBlok:
    def get_dns_result(self):
        return self.result

    def preflight(self, hostnames: list[str], ip_addresses: list[str]):
        self.result = DNSResult(hostnames=hostnames, ip_addresses=ip_addresses)

    def build(self, cwd):
        pass



from pathlib import Path
import traceback
from rich import get_console
import yaml
from blok.diff import compare_structures
from blok.io.read import create_structure_from_files_and_folders
from blok.io.write import create_files_and_folders
from blok.registry import BlokRegistry
from blok.errors import (
    DependencyNotFoundError,
    TooManyBlokFoundError,
    BlokBuildError,
    BlokInitializationError,
)
from blok.tree.models import YamlFile
from blok.utils import get_cleartext_deps, get_prepended_values, remove_empty_dicts
from blok.models import NestedDict
from blok.blok import ExecutionContext, InitContext
import rich_click as click
from rich.tree import Tree
from blok.render.tree import construct_file_tree, construct_diff_tree
from blok.render.panel import create_welcome_pane, create_dependency_resolutions_pane
import os
from blok.blok import Blok
from typing import List, Optional
from collections import OrderedDict
from blok.renderer import Renderer
import subprocess
from blok.renderer import Renderer, Panel


def secure_path_combine(x: Path, y: Path) -> Path:
    # Resolve the combined path

    mother_path = x.resolve()

    combined_path = (mother_path / y).resolve()

    # Check if the combined path is within the mother path
    if mother_path in combined_path.parents or mother_path == combined_path:
        return combined_path
    else:
        raise ValueError(
            "The user-defined path traverses out of the mother path. {} but requested {}".format(
                list(combined_path.parents), mother_path
            )
        )


@click.pass_context
def entrypoint(
    ctx: click.Context,
    registry: BlokRegistry,
    renderer: Renderer,
    blok_file_name: str,
    **kwargs,
):
    renderer.render(
        Panel(
            "Welcome to Blok!", title="Lets install this project", style="bold magenta"
        )
    )

    path = Path(kwargs.pop("path"))
    yes = kwargs.pop("yes", False)

    select = kwargs.pop("select", [])

    install_commands = ctx.obj.get("install_commands", {})

    selected_commands = []

    if select:
        print("Selecting commands")
        for key in select:
            if key in install_commands:
                selected_commands.append(install_commands[key])
            else:
                raise click.ClickException(f"Command with key {key} not found")

    else:
        selected_commands = list(install_commands.values())

    if not selected_commands:
        raise click.ClickException("No install commands found in blok configuration")

    for command in selected_commands:
        if yes or renderer.confirm(
            "Run install command?: " + " ".join(command["command"])
        ):
            rel_path = secure_path_combine(path, Path(command["cwd"]))
            subprocess.run(" ".join(command["command"]), shell=True, cwd=rel_path)



from blok.bloks.netifaces import NetifacesBlok
from blok.bloks.dns_prompt import PromptDNSBlok
from blok.bloks.vscode import VsCodeBlok


def get_bloks():
    return [NetifacesBlok(), VsCodeBlok(), PromptDNSBlok()]



from dataclasses import dataclass
import typing as t


@dataclass
class Panel:
    content: str
    title: str
    style: str = "bold magenta"
    expand: bool = False


Scene = Panel


@t.runtime_checkable
class Renderer(t.Protocol):
    def render(self, scene: Scene) -> None: ...

    def ask(self, question: str) -> str: ...

    def confirm(self, question: str) -> bool: ...



from blok.cli.create import make_cli
from blok.cli.magic_registry import MagicRegistry
from blok.renderers.click import RichRenderer
from rich import get_console
import rich_click as click

magic = MagicRegistry("__blok__")

magic.load_magic()

renderer = RichRenderer(get_console())

WELCOME_MESSAGE = """"
Welcome to blok! A tool for building and managing docker-compose projects.

Blok utilized your locally installed projects to build and manage docker-compose projects.
Projects can register bloks into the blok registry using the __blok__ magic method.
For more information, visit [link=https://arkitekt.live/bloks]https://arkitekt_next.live/bloks[/link]

"""

LOGO = r"""
 ____   _       ___   __  _ 
|    \ | |     /   \ |  |/ ]
|  o  )| |    |     ||  ' / 
|     || |___ |  O  ||    \ 
|  O  ||     ||     ||     \
|     ||     ||     ||  .  |
|_____||_____| \___/ |__|\_|                             
"""

ERROR_EPILOGUE = "To find out more, visit [link=https://arkitekt.live/bloks]https://arkitekt.live/bloks[/link]"


click.rich_click.HEADER_TEXT = LOGO + "\n"
click.rich_click.ERRORS_EPILOGUE = ERROR_EPILOGUE
click.rich_click.USE_RICH_MARKUP = True


cli = make_cli(magic, renderer)



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass
import socket
from dataclasses import dataclass, field
from typing import Set


@service("io.blok.vscode")
class VsCodeService(Protocol):
    def register_task() -> None:
        pass



import typing as t
from dataclasses import dataclass


@dataclass
class Option:
    subcommand: str
    show_default: t.Union[bool, str, None] = None
    prompt: t.Union[bool, str] = False
    confirmation_prompt: t.Union[bool, str] = False
    prompt_required: bool = True
    hide_input: bool = False
    is_flag: t.Optional[bool] = None
    flag_value: t.Optional[t.Any] = None
    count: bool = False
    allow_from_autoenv: bool = True
    help: t.Optional[str] = None
    hidden: bool = False
    show_choices: bool = True
    show_envvar: bool = False
    type: t.Optional[t.Union[t.Any, t.Any]] = None
    required: bool = False
    default: t.Optional[t.Union[t.Any, t.Callable[[], t.Any]]] = None
    callback: t.Optional[t.Callable[[t.Any, t.Any, t.Any], t.Any]] = None
    nargs: t.Optional[int] = None
    multiple: bool = False
    metavar: t.Optional[str] = None
    expose_value: bool = True
    is_eager: bool = False
    shell_complete: t.Optional[
        t.Callable[
            [t.Any, t.Any, str],
            t.Union[t.List[t.Any], t.List[str]],
        ]
    ] = None


class NestedDict(dict):
    def set_nested(self, *keys_and_value):
        *keys, value = keys_and_value
        d = self  # Reference to the current level of the nested dictionary

        # Iterate through all keys except the last one
        for k in keys[:-1]:
            if k not in d or not isinstance(d[k], dict):
                d[k] = NestedDict()
            d = d[k]

        # Merge with existing if it's a dict, otherwise set the value
        last_key = keys[-1]
        if isinstance(d.get(last_key), dict) and isinstance(value, dict):
            d[last_key].update(value)
        else:
            d[last_key] = value



import json
import os
from pathlib import Path
from blok.tree.models import JSONFile, YamlFile, Repo
import yaml
import subprocess


def create_structure_from_files_and_folders(base_path, omit: list[str] | None = None):
    if omit is None:
        omit = []
    structure = {}

    def get_repo_url(repo_path):
        try:
            result = subprocess.run(
                ["git", "-C", repo_path, "config", "--get", "remote.origin.url"],
                capture_output=True,
                text=True,
                check=True,
            )
            return result.stdout.strip()
        except subprocess.CalledProcessError as e:
            print(f"Failed to get repo URL for {repo_path}: {e}")
            return None

    def process_directory(current_path, current_dict):
        for root, dirs, files in os.walk(current_path):
            rel_path = Path(root).relative_to(base_path)
            if ".git" in dirs:
                # Handle git repository
                repo_url = get_repo_url(root)
                if repo_url:
                    temp_dict = current_dict
                    for part in rel_path.parts:
                        temp_dict = temp_dict.setdefault(part, {})
                    temp_dict[".git"] = Repo(value=repo_url)
                continue

            # Update current_dict to point to the correct level in the structure
            temp_dict = current_dict
            for part in rel_path.parts:
                temp_dict = temp_dict.setdefault(part, {})

            for dir_name in dirs:
                temp_dict[dir_name] = {}

            for file_name in files:
                if omit and file_name in omit:
                    continue

                file_path = Path(root) / file_name

                try:
                    with file_path.open("rb") as file:
                        content = file.read()

                    if file_name.endswith(".yaml") or file_name.endswith(".yml"):
                        content_decoded = content.decode()
                        yaml_content = yaml.safe_load(content_decoded)
                        temp_dict[file_name] = YamlFile(**yaml_content)
                    elif file_name.endswith(".json"):
                        content_decoded = content.decode()
                        json_content = json.loads(content_decoded)
                        temp_dict[file_name] = JSONFile(**json_content)
                    else:
                        try:
                            content_decoded = content.decode()
                            temp_dict[file_name] = content_decoded
                        except UnicodeDecodeError:
                            temp_dict[file_name] = content
                except Exception as e:
                    print(f"Failed to read {file_path}: {e}")

            # Recursively process subdirectories
            for dir_name in dirs:
                process_directory(Path(root) / dir_name, temp_dict[dir_name])
            break  # Prevent further recursion by os.walk since we manually process subdirectories

    process_directory(base_path, structure)
    return structure



from blok.renderer import Renderer, Panel
from rich import panel
import rich_click as click


class RichRenderer(Renderer):
    def __init__(self, console):
        self.console = console

    def render(self, scene: Panel) -> None:
        if isinstance(scene, Panel):
            self.console.print(
                panel.Panel(
                    scene.content,
                    title=scene.title,
                    style=scene.style,
                    expand=scene.expand,
                )
            )
            return

        raise NotImplementedError("Scene type not supported")

    def confirm(self, question: str) -> bool:
        return click.confirm(question)

    def ask(self, question: str) -> str:
        return click.prompt(question)



from blok.registry import BlokRegistry, Blok
from blok.utils import check_protocol_compliance
import importlib
import os
import traceback
import pkgutil
import typing as t


def load_and_call_get_blocks(module_name, rekuest_path, magic_path):
    try:
        spec = importlib.util.spec_from_file_location(
            f"{module_name}.{magic_path}", rekuest_path
        )
        rekuest_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(rekuest_module)
        if hasattr(rekuest_module, "get_bloks"):
            bloks = rekuest_module.get_bloks()
            assert isinstance(bloks, list), "get_bloks must return a list"
            return bloks
        else:
            print(f"No get_bloks function in {module_name}.{magic_path}")
    except Exception as e:
        print(f"Failed to call get_bloks for {module_name}: {e}")
        traceback.print_exc()

    return []


def check_and_import_magic(magic_path: str):
    registered_bloks = []

    # Check local modules in the current working directory
    current_directory = os.getcwd()
    for item in os.listdir(current_directory):
        item_path = os.path.join(current_directory, item)
        if os.path.isdir(item_path) and os.path.isfile(
            os.path.join(item_path, "__init__.py")
        ):
            rekuest_path = os.path.join(item_path, f"{magic_path}.py")
            if os.path.isfile(rekuest_path):
                registered_bloks.extend(
                    load_and_call_get_blocks(item, rekuest_path, magic_path)
                )

    # Check installed packages
    for _, module_name, _ in pkgutil.iter_modules():
        try:
            module_spec = importlib.util.find_spec(module_name)
            if module_spec and module_spec.origin:
                rekuest_path = os.path.join(
                    os.path.dirname(module_spec.origin), f"{magic_path}.py"
                )
                if os.path.isfile(rekuest_path):
                    registered_bloks.extend(
                        load_and_call_get_blocks(module_name, rekuest_path, magic_path)
                    )

        except Exception as e:
            print(f"Failed to call get_bloks for installed package {module_name}: {e}")
            traceback.print_exc()

    return registered_bloks


class MagicRegistry(BlokRegistry):
    def __init__(self, magic_path: str, **kwargs):
        super().__init__(**kwargs)
        self.magic_path = magic_path

    def load_magic(self):
        registered_blocks = check_and_import_magic(self.magic_path)
        for blok in registered_blocks:
            self.add_blok(blok)



from dataclasses import asdict
import typing as t
from blok.blok import Blok
from blok.utils import (
    check_allowed_module_string,
    check_protocol_compliance,
)
import importlib


def lazy_load_blok(path) -> t.Tuple[str, Blok]:
    # lazily loading a command, first get the module name and attribute name
    import_path = path

    blok_name = "_".join(import_path.split("."))
    modname, cmd_object_name = import_path.rsplit(".", 1)
    # do the import
    mod = importlib.import_module(modname)
    # get the Command object from that module
    cmd_object = getattr(mod, cmd_object_name)
    # check the result to make debugging easier

    blok = cmd_object()
    check_protocol_compliance(blok, Blok)

    return blok_name, blok


class BlokRegistry:
    def __init__(self, strict: bool = False):
        self.services: t.Dict[str, Blok] = {}
        self.bloks: t.Dict[str, Blok] = {}
        self.dependency_resolver = {}
        self.strict = strict
        self.meta = {}

    def load_modules(self, modules: t.List[str]):
        for module in modules:
            self.load_module(module)

    def get_blok(self, blok_key: str):
        try:
            return self.bloks[blok_key]
        except KeyError:
            raise KeyError(f"Could not find blok with key {blok_key}")

    def get_bloks_for_dependency(
        self,
        identifier: str,
    ):
        blok_keys = self.get_module_name(identifier)
        return [self.get_blok(blok_key) for blok_key in blok_keys]

    def load_module(self, module: str, with_key: t.Optional[str] = None):
        key, blok = lazy_load_blok(module)
        self.add_blok(with_key or key, blok)

    def add_blok(self, blok: Blok):
        check_protocol_compliance(blok, Blok)
        meta = blok.get_blok_meta()

        if meta.name in self.bloks:
            if self.strict:
                raise KeyError(
                    f"Blok {meta.name} already exists. Cannot register it twice. Choose a different name."
                )
        else:
            self.dependency_resolver.setdefault(meta.service_identifier, []).append(
                meta.name
            )

        self.meta[meta.name] = meta

        self.bloks[meta.name] = blok

    def get_module_name(self, identifier):
        return self.dependency_resolver[identifier]

    def get_click_options(self):
        import rich_click as click

        integrated_options = []

        for blok_key, blok in self.bloks.items():
            for option in blok.get_options():
                params = asdict(option)

                subcommand = params.pop("subcommand")
                assert subcommand, "subcommand is required"
                assert check_allowed_module_string(
                    subcommand
                ), "subcommand must be a valid python variable name"

                integrated_option = click.option(
                    f"--{blok_key.replace('_', '-')}-{subcommand.replace('_', '-')}",
                    envvar=f"{blok_key.upper()}_{subcommand.upper()}",
                    **params,
                )

                integrated_options.append(integrated_option)

        return integrated_options



from typing import List, runtime_checkable, Protocol
import typing as t
from dataclasses import dataclass


@dataclass
class ServiceMeta:
    name: t.Optional[str]
    identifier: str
    description: t.Optional[str]


class Service(Protocol):
    def get_blok_service_meta(self) -> ServiceMeta:
        raise NotImplementedError("This method must be implemented by the subclass")


def service(
    identifier: str, description: t.Optional[str] = None, name: t.Optional[str] = None
):
    def decorator(cls):
        cls.__blok_meta__ = ServiceMeta(
            identifier=identifier, description=description or "", name=name
        )
        cls.__is_service__ = True

        if not hasattr(cls, "get_blok_service_meta"):
            cls.get_blok_service_meta = classmethod(lambda self: self.__blok_meta__)

        try:
            return runtime_checkable(cls)
        except TypeError as e:
            raise TypeError(
                f"Could not create Blok Service from {cls}. Blok services need to inherit from 'typing.Protocol'"
            ) from e

    return decorator
